;; Generated package description from gptel.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "gptel" "0.9.9.4.0.20260503.114603" "Interact with ChatGPT or other LLMs" '((emacs "27.1") (transient "0.7.4") (compat "30.1.0.0")) :commit "f915a8bba57fcd4fe7d92bf2bbcec17e5bbfa25b" :authors '(("Karthik Chikmagalur" . "karthik.chikmagalur@gmail.com")) :maintainer '("Karthik Chikmagalur" . "karthik.chikmagalur@gmail.com") :keywords '("convenience" "tools") :url "https://github.com/karthink/gptel")
